# toy program: Kanji Quiz - Kanji/English quiz and guessing
# game.
#
# Inspired by Philip Brown's jdrill: botlhole.com
# 
# Uses Kanji/English dictionaries extracted from kanjidic2:
# http://nihongo.monash.edu/kanjidic2/
#
# Most of the original xml entries have been stripped, converted
# to csv entries, then to lists organized by Kanji difficulty grade
#
# Grades range from 1 to 10 (with no 7) and null (ungraded and
# most difficult - which we will call grade 0)
#
# The grade naming - and missing 7 - make it tough to automate handling
# with lists and foreach.  First coding will deal with each explicitly.


package require Tk


# 0.8.0  2018.10.06
# 1st ~fully functional code. Needs code clean up and removal of diag prints.
#
# 0.8.1  2018.10.07
# Prettied up the code a bit and removed dprint stuff.
# Added version and date to "about"
#
# 0.8.2  2018.10.10
# Added code to detect our directory.  Necessary for portability when
# including the dictionary files via FreeWrap (Windows) or
# Androwish SDK (Android)
#
# 0.8.3  2018.10.11
# Added a half-assed line wrap proc. Some of the English meanings ca
# have quite long entries - usually toward the end of the list.  Try
# to wrap to a " {"
#
# Add a leave event after button presses. Without it when combined
# with Androwish SDK and run on Android a pressed button stays white
# until some other button is pressed. See below...
#
# The leave event caused as much trouble as it fixed.  Try just fiddling
# the active background color.  Will be constant for quiz name and quiz
# word buttons but will have to be adjusted on the fly for guess buttons.

set quizversion "0.8.3"
set quizdate "2018.10.11"

# If we use hard-coded unicode in the few places where we need it
# internally we don't have to force encoding for the entire program
# which would have to be done externally

# set quiznamejapanese  "漢字クイズ"

set quiznamejapanese "\u6F22\u5B57\u30AF\u30A4\u30BA"
set quiznameenglish "Kanji Quiz"

# Quiz language mode: kanji = Kanji quiz word, English guesses
# english = reverse
#
# YES, I REALIZE KANJI ISN'T A LANGUAGE.  So sue me...

set quizlanguage "kanji"

# check the below for appropraite noun/verb/...

set optionsjapanese "\u30AA\u30D7\u30B7\u30E7\u30F3"
set helpjapanese "\u624B\u4F1D"
set aboutjapanese "\u60C5\u5831"
set exitjapanese "\u51FA\u53E3"

# Default option checkbutton settings: grade 1 selected.
# Grade 7 is always OFF (doesn't exist)

set boxgrade1 1
set boxgrade7 0

# Possibly redundant but default available grades to #1

set availablegrades [ list "grade1" ]

# Colors and other handy literals - all in one place

set borderwidth 10

set bordercolor lightblue
set quiznamecolor indianred
set quizwordcolor seagreen
set guesswordcolor lightcoral
set guessrightcolor seagreen
set guesscheatcolor seagreen
set guesswrongcolor red
set dupcolor yellow
set menucolor gray93

set quiztextfont helvetica
set quiztextsize 18
set guesstextfont helvetica
set guesstextsize 16
set messagetextfont helvetica
set messagetextsize 10
set checktextfont helvetica
set checktextsize 10

set wraplimit 50


# Canned messages for the menu buttons:

set aboutmessage\
"Kanji Quiz - $quizversion - $quizdate\n\
A toy tcl/tk program to (hopefully)\
help with learning Kanji.\n
Project home page: http://www.beezmo.com/geezblog/?p=1842\n
Kanji/English dictionaries are derived from the kanjidic2 dictionary available at\
http://nihongo.monash.edu/kanjidic2/"

set helpmessage\
"Select grade(s) - difficulty and number of possible characters
in the Options menu.\n
Clicking on the \"Kanji Quiz <-> \u6F22\u5B57\u30AF\u30A4\u30BA\" button toggles
between Kanji quiz word/English choices, and v/v.\n
Clicking on the quiz word itself will \"cheat\" by highlighting
the correct answer.\n
Click on one of the five possbile answers to select your answer.
It will turn red if wrong.  You can keep trying. If correct it
will turn green and things will move to the next random quiz."


# Canned message for the occasions when rand produces duplicates:

set dupmessage\
"Well, this is embarrassing.  In spite of our best efforts,\
the random guess thingy has produced duplicates.\n
It's harmless."


# list of grades

set grades [ list "grade0" "grade1" "grade2" "grade3" "grade4" "grade5" \
	"grade6" "grade8" "grade9" "grade10" ]

# list of matching dictionary file names

set dictfiles [ list "KanjiGradeNullList" "KanjiGradeOneList"  "KanjiGradeTwoList" \
	"KanjiGradeThreeList" "KanjiGradeFourList" "KanjiGradeFiveList" "KanjiGradeSixList" \
	"KanjiGradeEightList" "KanjiGradeNineList" "KanjiGradeTenList" ]


	
# Set the list of available grades
# Will default to grade 1 if none set - when doing that turn on the grade1
# check button variable so that check button will be selected
#
# There is no grade 7.


proc setAvailableGrades {} {

    global availablegrades grades
    global  boxgrade0  boxgrade1 boxgrade2 boxgrade3 boxgrade4 boxgrade5 \
    	boxgrade6 boxgrade8 boxgrade9 boxgrade10

    set availablegrades [ list ]
    foreach gradex $grades {
	if { [ set "box$gradex" ] == 1 } { lappend availablegrades "$gradex" }
    }
    if { [ llength $availablegrades ] == 0 } {
	
	set boxgrade1 1
	set availablegrades [ list "grade1" ]
    }
    return 0
}


# Pop-up text window.  Call with text, background color, maybe later font,
# font size, and window size
#
# When done wait here until window is closed so we don't exit some calling code
# without displaying the message
#
# We don't know the calling context to try to anchor to main GUI window
# if it exists - this doesn't work very well if it's an error window early
# in the process. Gets located relative to screen, not GUI - even though
# it should...

proc popMessage { msgtext msgcolor } {

    global messagetextfont messagetextsize

    if { [ winfo exists .popmsg ] } { return }

    toplevel .popmsg -bg $msgcolor
    wm title .popmsg "Kanji Quiz"

    if { [ winfo exists .kq ] } {
	wm transient .popmsg .kq
	wm geometry .popmsg +[ expr { [ winfo x . ] + 30  } ]+[ expr { [winfo y . ] + 30 } ]
    }
    grab .popmsg
    message .popmsg.msg -text $msgtext -aspect 500 -font "$messagetextfont $messagetextsize" \
    	-bg $msgcolor
    button .popmsg.okb -text OK -command {destroy .popmsg ; return 0}

    grid .popmsg.msg
    grid .popmsg.okb

    tkwait window .popmsg

    return 0
}


# Try to wrap long English entries.  Check against wrap limit and back
# up from there looking for a space and brace combo. Insert newline and
# replace space and brace (indented)
#
# Remember that linetowrap is a string, not a list

proc fishWrap { linetowrap } {

    global wraplimit

    if { [ string length $linetowrap ] > $wraplimit } {

	set spaceindex [ string last " \{" $linetowrap $wraplimit ]

	return [ string replace $linetowrap $spaceindex [ expr { $spaceindex + 1 } ] "\n\{" ]

    } else {
	return $linetowrap
    }

}


# Read the ditionary files (lists) into the grade lists
# Set the encoding of the incoming files - not sure this is necessary
# for this project but it allows printing kanji in a windows terminal
#
# It creates the global grades on the fly
#
# I've had trouble popping up an error window on top of main - since
# the main window may not be established when this is run.
#
# Return the name of the missing file if not found, or 0 if things are OK.
#
# Updated to detect our directory - required since full path to files
# needed in FreeWrap and Androwish SDK version.

proc readDictionaries {} {

    global grades dictfiles

    set ourdir [ file dirname [ info script ] ]

    foreach gradex $grades dictfile $dictfiles {

	if { [ file exists $ourdir/$dictfile ] != 1 } {
	    return $dictfile
	}
	set infile [ open $ourdir/$dictfile r ]
	fconfigure $infile -encoding utf-8

	# does this actually create global lists on the fly?
	global list $gradex
	set $gradex [ gets $infile ]
	close $infile
    }
    return 0
}


# get a random element from a list

proc getRandomElement { listarg } {

    return [ lindex $listarg [  expr { int( rand() * [ llength $listarg ] ) } ] ]

    # returns lindex result
}


# Pop up an "options" checkbutton box that lets the user set available grades.
# Make it top-level so size isn't limited by main GUI
#
# Will do each by hand since there's no 7 - maybe should be doing foreach
# from the grade list?
#
# Associated variable defaults to the end of the button path name if not
# specified and are automatically global.

proc popButtonBox {} {

    global checktextfont checktextsize

    if { [ winfo exists .popboxes ] } { return }

    toplevel .popboxes
    wm title .popboxes "Kanji Quiz"
    wm transient .popboxes .kq
    wm geometry .popboxes +[ expr { [ winfo x . ] + 30  } ]+[ expr { [winfo y . ] + 30 } ]
    grab .popboxes

    checkbutton .popboxes.boxgrade1 -font "$checktextfont $checktextsize" -text "Grade 1 (easiest)" 
    checkbutton .popboxes.boxgrade2 -font "$checktextfont $checktextsize" -text "Grade 2"
    checkbutton .popboxes.boxgrade3 -font "$checktextfont $checktextsize" -text "Grade 3"
    checkbutton .popboxes.boxgrade4 -font "$checktextfont $checktextsize" -text "Grade 4" 
    checkbutton .popboxes.boxgrade5 -font "$checktextfont $checktextsize" -text "Grade 5" 
    checkbutton .popboxes.boxgrade6 -font "$checktextfont $checktextsize" -text "Grade 6" 
    checkbutton .popboxes.boxgrade7 -font "$checktextfont $checktextsize" -text "(No Grade 7)"\
    	-state disabled
    checkbutton .popboxes.boxgrade8 -font "$checktextfont $checktextsize" -text "Grade 8" 
    checkbutton .popboxes.boxgrade9 -font "$checktextfont $checktextsize" -text "Grade 9" 
    checkbutton .popboxes.boxgrade10 -font "$checktextfont $checktextsize" -text "Grade 10" 
    checkbutton .popboxes.boxgrade0 -font "$checktextfont $checktextsize" -text "Ungraded (hardest)"

    button .popboxes.gckapply -text Apply -command { destroy .popboxes ; setAvailableGrades }

    pack .popboxes.boxgrade1 .popboxes.boxgrade2 .popboxes.boxgrade3 \
    .popboxes.boxgrade4 .popboxes.boxgrade5 .popboxes.boxgrade6 .popboxes.boxgrade7 \
    .popboxes.boxgrade8 .popboxes.boxgrade9 .popboxes.boxgrade10 .popboxes.boxgrade0 \
    .popboxes.gckapply -anchor w -in .popboxes

    tkwait window .popboxes
    return 0
}


# Launch next quiz.  We have to do this as a separate proc since
# using "after" and sleeping prevents things from happening, including
# possibly this display of the "correct guess" color.

proc nextQuiz {} {

	getFiveAtRandom
	pickQuizWord
	restoreColors
	showQuizAndGuesses

}


# Come here when a guess button is pressed.
# If the incoming index matches the quiz word index turn green.
# Otherwise, turn red.
#
# Will need to add to this and after green pause a second or
# two and load a new quiz word and guesses. AND restore the colors.
#
# Change active background color along with guess color to try to
# avoid issues with Android buttons sticking at white, and Linux
# buttons turning - and staying - white while in mouse over.


proc guessPressed { buttonindex } {

    global quizwordindex guesswrongcolor guessrightcolor

    if { $buttonindex == $quizwordindex } {
	.kq.guessbutton$buttonindex configure -bg $guessrightcolor \
	-activebackground $guessrightcolor

	# We do it this way (immediate after return w/nextQuiz launched
	# later since just "after 1000" can suspend the app before the
	# "good guess" color gets displayed and the button just stays
	# white during the delay.  MIGHT BE a better way. ????
	
	after 2000 nextQuiz
    } else {
	.kq.guessbutton$buttonindex configure -bg $guesswrongcolor \
	-activebackground $guesswrongcolor
    }
    return 0
}


# When the big quiz name button is pressed, toggle between Japanese
# and English quiz.  global quizlanguage: 0 = Kanji, 1 = English

proc toggleLanguage {} {

    global quizlanguage

    if { [ string match $quizlanguage "kanji" ] } {
        set quizlanguage "english"
    } else {
	set quizlanguage "kanji"
    }
    return 0
}


# Get one item at random from available grades. The calling program will
# check that we don't come up with duplicates and call again if needed
#
# First get avaialable grades and pick one at random.  Then from that
# grade pick an item at random.

proc getOneAtRandom {} {

    global availablegrades
    global grade0 grade1 grade2 grade3 grade4 grade5 grade6 grade8 grade9 grade10

    set gradex [ getRandomElement $availablegrades ]
    set itemx [ getRandomElement [ subst $$gradex ] ]
    return $itemx
}


# Get five random (and different) items from available grades.  Will call
# getOne... and check for dups.
#
# Should they be five separate named lists or five lists that are elements
# of a larger list ???  what will be easiest to have as button commands?
#
# We'll check for dups as we go.  If a given guess is a dup we'll try
# again - maybe more than once.  TBD.  the once in a blue moon
# dup shouldn't cause much concern...

proc getFiveAtRandom {} {

    global fiveguesses dupmessage dupcolor

    set fiveguesses [ list ]
    for { set index 0 } { $index <  5 } { incr index } {
	set oneguess [ getOneAtRandom ]
	foreach guesscheck $fiveguesses {
	    
	    # Check the kanji characters for a duplicate. We will try twice
	    # more and if those fail we pop a warning/explanation. Lame.
	    # Could implement a counter but easier to just unroll...

	    if { [ lindex $oneguess 0 ] == [ lindex $guesscheck 0 ] } {
		set oneguess [ getOneAtRandom ]
	    }
	    if { [ lindex $oneguess 0 ] == [ lindex $guesscheck 0 ] } {
		set oneguess [ getOneAtRandom ]
	    }

	    if { [ lindex $oneguess 0 ] == [ lindex $guesscheck 0 ] } {
		popMessage $dupmessage $dupcolor
	    }
	}
	lappend fiveguesses $oneguess
    }
    return 0
}


# Select which of the five guess words to use as the quiz word
# at random.  We're hardcoding the number of guess words at five for now

proc pickQuizWord {} {

    global quizwordindex

    set quizwordindex [ expr { int( rand() * 5 ) } ]
    return 0
}


# When the quiz word button is pressed, show the cheat answer

proc showCheat {} {

    global quizwordindex guesscheatcolor

    .kq.guessbutton$quizwordindex configure -bg  $guesscheatcolor
    return 0
}


# Restore the GUI colors. We could rebuild from zero but try this
# for now.

proc restoreColors {} {

    global guesswordcolor
       
    for { set index 0 } { $index <  5 } { incr index } {
	.kq.guessbutton$index configure -bg $guesswordcolor
    }
}


# Actually display the quiz word and guesses (or at least prep their
# entries - TBD) in Kanji or English depending on global switch
#
# Use join to strip curly braces from multi-word entries - and put them
# back for all entries.
#
# Same switch might toggle the kanji quiz <-> button side to side
#
# !!! Maybe a different approach: proc to produce quiz word and a
# different one to produce guess word based on its index into 5
# guesses - in the correct language.  might be handy to use in the
# proc that configures the board.
#
# !!!!!!! do we set colors again here?  Or only here?

proc showQuizAndGuesses {} {

    # This will be simpler if we split the whole thing into two
    # branches depending on the language mode.
    #
    # When we arrive here we have fiveguesses and a quizword index
    # into fiveguesses
    #
    # Need to restore colors here since previous active background
    # colors can flash during button press.

    global quizlanguage fiveguesses quizwordindex guesswordcolor

    if { $quizlanguage == "kanji" } {

	# Mode: kanji quiz word, English guesses

	.kq.quizword configure -text [ lindex [ lindex $fiveguesses $quizwordindex ] 0 ]
	for { set index 0 } { $index <  5 } { incr index } {
	    .kq.guessbutton$index configure \
	    	-text [ fishWrap \
		    "{[join [ lrange [ lindex $fiveguesses $index ]  2 end ] "} {" ]}" ] \
		    -bg $guesswordcolor -activebackground $guesswordcolor
	}
    } else {

	# Mode: English quiz word, kanji guesses

	.kq.quizword configure \
		-text [ fishWrap \
		    "{[join [ lrange [ lindex $fiveguesses $quizwordindex ]  2 end ] "} {" ]}" ]
	for { set index 0 } { $index <  5 } { incr index } {
	    .kq.guessbutton$index configure \
	    	-text "[ lindex [ lindex $fiveguesses $index ]  0 ]" \
		-bg $guesswordcolor -activebackground $guesswordcolor
	}
    }
    return 0
}


# Quiz button presses so toggle language and redisplay.
# Could combine commands in button line but this may give
# us more flexibility as this evolves.

proc quizPressed {} {

    toggleLanguage
    showQuizAndGuesses
}


# Assemble the structure of the main GUI
#
# Sets sizes and colors but not text in quiz word or guesses
#
# Set up active background color to match background color to
# try to solve problem in Android where buttons stay white until
# something else pressed.

proc rebuildQuizGUI {} {

    global quiztextfont quiztextsize guesstextfont guesstextsize

    global quiznameenglish quiznamejapanese fiveguesses borderwidth bordercolor \
    	quiznamecolor quizwordcolor guesswordcolor menucolor
	
    if { [winfo exists .kq] } { destroy .kq }
    set rowcount 0

    frame .kq -borderwidth $borderwidth -bg $bordercolor
    wm title . "Kanji Quiz"

    button .kq.options -text Options -command popButtonBox
    button .kq.help -text Help -command { popMessage $helpmessage $menucolor }
    button .kq.about -text About -command { popMessage $aboutmessage $menucolor }
    button .kq.exit -text Exit -command exit

    button .kq.quizname -text "$quiznameenglish \u2b0c $quiznamejapanese" \
    	-font "$quiztextfont $quiztextsize" -bg $quiznamecolor \
	-activebackground $quiznamecolor -command quizPressed

    button .kq.quizword -font "$quiztextfont $quiztextsize" \
    	-bg $quizwordcolor -activebackground $quizwordcolor -command  showCheat

    frame .kq.separator0 -height $borderwidth -bg $bordercolor
    frame .kq.separator1 -height $borderwidth -bg $bordercolor
    frame .kq.separator2 -height $borderwidth -bg $bordercolor

    grid .kq.options -row 0 -column 0 -sticky nsew
    grid .kq.help -row 0 -column 1 -sticky nsew
    grid .kq.about -row 0 -column 2 -sticky nsew
    grid .kq.exit -row 0 -column 3 -sticky nsew ; incr rowcount

    grid .kq.separator0 -row 1 -column 0 -columnspan 4 ; incr rowcount
    grid .kq.quizname -row 2 -column 0 -columnspan 4 -sticky nsew ; incr rowcount
    grid .kq.separator1 -row 3 -column 0 -columnspan 4 ; incr rowcount
    grid .kq.quizword -row 4 -column 0 -columnspan 4 -sticky nsew ; incr rowcount
    grid .kq.separator2 -row 5 -column 0 -columnspan 4 ; incr rowcount
    
    for { set index 0 } { $index <  5 } { incr index } {
	set buttonnumber [expr { $index + $rowcount } ]
	button .kq.guessbutton$index -font "$guesstextfont $guesstextsize" \
		-bg $guesswordcolor -activebackground $guesswordcolor \
		-command "guessPressed $index"
	grid .kq.guessbutton$index -row $buttonnumber -column 0 -columnspan 4 -sticky nsew
    }
    grid .kq
    return 0
}


#showtime....


rebuildQuizGUI

set readdictreturn [ readDictionaries ]
if { $readdictreturn != 0 } {

    popMessage "$quiznameenglish: Dictionary file $readdictreturn not found!" red
    exit

}

getFiveAtRandom
pickQuizWord
showQuizAndGuesses






